#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
Developers: use this code to develop and test validate.py
'''


import os
# disable HDF5 library/header mismatch warning for now
# Headers are 1.8.15, library is 1.8.16
# os.environ['HDF5_DISABLE_VERSION_CHECK'] = '2'
import validate
import finding


PKG_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(PKG_DIR, 'data')
TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, 'writer_1_3.hdf5')
# TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, 'writer_2_1.hdf5')
# TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, '02_03_setup.h5')
# TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, 'chopper.nxs')
# TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, 'scan101.nxs')
# TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, 'compression.h5')
# TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, 'Data_Q.h5')
# TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, 'USAXS_flyScan_GC_M4_NewD_15.h5')

# these two files for testing contain non-standard items and NeXus errors
# TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, 'draft_1D_NXcanSAS.h5')
# TEST_DATA_FILE = os.path.join(TEST_DATA_DIR, 'draft_2D_NXcanSAS.h5')

validator = validate.Data_File_Validator(TEST_DATA_FILE)
validator.validate()


# report the findings from the validation
#  finding.SHOW_ALL        finding.SHOW_NOT_OK        finding.SHOW_ERRORS
show_these = finding.SHOW_ALL
print 'Validation findings'
print ':file: ' + os.path.basename(validator.fname)
print ':validation results shown: ', ', '.join(sorted(map(str, show_these)))
print validator.report_findings(show_these)

print 'summary statistics'
print validator.report_findings_summary()

# print 'NeXus classpath full map'
# print validator.report_classpath()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
Developers: use this code to develop and test nxdlstructure.py
'''


import os
import sys
import cache
import nxdlstructure


def show_NXDL_structure(category, NX_class):
    ''''find the directory of this python file'''
    qset = cache.qsettings()
    BASEDIR = qset.nxdl_dir()
    if not os.path.exists(BASEDIR):
        raise IOError('no NXDL cache found, need to create it with an *update*')

    nxdl = os.path.join(BASEDIR, category, NX_class + '.nxdl.xml')
    
    if len(sys.argv) == 1:
        sys.argv.append(nxdl)
    elif len(sys.argv) > 1:
        sys.argv[1] = nxdl
    
    nxdlstructure.main()


def get_NXDL_dictionary():
    nxdl_dict = nxdlstructure.get_NXDL_specifications()
    print len(nxdl_dict)


# show_NXDL_structure('base_classes', 'NXentry')
# show_NXDL_structure('base_classes', 'NXcrystal')
# show_NXDL_structure('base_classes', 'NXdata')
# show_NXDL_structure('base_classes', 'NXobject')
# show_NXDL_structure('applications', 'NXarchive')
# show_NXDL_structure('applications', 'NXsas')
# show_NXDL_structure('applications', 'NXarpes')
# show_NXDL_structure('contributed_definitions', 'NXmagnetic_kicker')
show_NXDL_structure('contributed_definitions', 'NXcanSAS')

# get_NXDL_dictionary()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
Interpret the NXDL rules (nxdl.xsd & nxdlTypes.xsd) into useful Python components

.. index:: NXDL

:definition: NXDL : NeXus Definition Language
:see: http://download.nexusformat.org/doc/html/nxdl.html

.. rubric:: Structure type coverage

These NXDL structures are parsed now by the code below:

* [x] attributeType
* [x] basicComponent
* [x] definitionType
* [x] definitionTypeAttr
* [x] dimensionsType
* [x] docType
* [x] enumerationType
* [x] fieldType
* [x] groupGroup
* [x] groupType
* [x] linkType
* [x] nonNegativeUnbounded
* [x] symbolsType
* [x] validItemName
* [x] validNXClassName
* [x] validTargetName
'''


import lxml.etree

import cache


PROGRAM_NAME = 'nxdl_rules'
NXDL_XML_NAMESPACE = 'http://definition.nexusformat.org/nxdl/3.1'
XMLSCHEMA_NAMESPACE = 'http://www.w3.org/2001/XMLSchema'
NAMESPACE_DICT = {'nx': NXDL_XML_NAMESPACE, 
                  'xs': XMLSCHEMA_NAMESPACE}


__group_parsing_started__ = False   # avoid a known recursion of group in a group


def strip_ns(ref):
    '''
    strip the namespace prefix from ``ref``
    
    :param str ref: one word, colon delimited string, such as *nx:groupGroup*
    :returns str: the part to the right of the last colon
    '''
    return ref.split(':')[-1]


class NxdlRules(object):
    '''
    Interpret the rules for the NXDL files
    
    These contents will be used as the defaults for each of the
    NXDL types (attribute, group, field, link) when parsing
    each NXDL specification.
    '''
    
    def __init__(self):
        self.ns = NAMESPACE_DICT
        nxdl_xsd = cache.get_nxdl_xsd()             # NXDL structures

        node_list = nxdl_xsd.xpath('xs:element', namespaces=self.ns)
        if len(node_list) != 1:
            self.raise_error(nxdl_xsd, 'wrong number of xs:element nodes found: ', len(node_list))

        self.nxdlTypes = self.parse_nxdlTypes()
        self.nxdl = NXDL_Root(node_list[0])
    
    def parse_nxdlTypes(self):
        db = {}
        nxdlTypes_xsd = cache.get_nxdlTypes_xsd()   # types of data and units
        for node in nxdlTypes_xsd.getroot():
            if isinstance(node, lxml.etree._Comment):
                pass
            elif node.tag.endswith('}annotation'):
                pass
            else:
                obj = NXDL_nxdlType(node)
                db[obj.name] = obj
        return db


class NXDL_nxdlType(object):
    '''
    one of the types defined in the file *nxdlTypes.xsd*
    '''
    
    def __init__(self, xml_obj):
        self.name = xml_obj.attrib.get('name')
        self.restriction = None
        self.union = None
        self.values = None
        
        for node in xml_obj:
            if isinstance(node, lxml.etree._Comment):
                pass
            elif node.tag.endswith('}annotation'):
                pass
            elif node.tag.endswith('}list'):
                self.values = map(strip_ns, [node.attrib['itemType'],])
            elif node.tag.endswith('}restriction'):
                self.restriction = strip_ns(node.attrib['base'])
                self.values = []
                for subnode in node:
                    if isinstance(node, lxml.etree._Comment):
                        pass
                    elif subnode.tag.endswith('}enumeration'):
                        self.values.append(subnode.attrib['value'])
            elif node.tag.endswith('}union'):
                self.union = map(strip_ns, node.attrib['memberTypes'].split())
            else:
                self.raise_error(node, 'unhandled tag=', node.tag)


class Mixin(object):
    '''
    common code for NXDL Rules classes below
    
    :param lxml.etree.Element xml_obj: XML element
    :param str obj_name: optional, default taken from ``xml_obj``
    :param dict ns_dict: optional, default taken from :data:`NAMESPACE_DICT`
    '''
    
    def __init__(self, xml_obj, obj_name=None, ns_dict=None):
        self.name = obj_name or xml_obj.attrib.get('name')
        self.ns = ns_dict or NAMESPACE_DICT
    
    def raise_error(self, node, text, obj):
        '''standard *ValueError* exception handling'''
        msg = 'line ' + str(node.sourceline)
        msg += ': ' + text + str(obj)
        raise ValueError(msg)
    
    def get_named_node(self, tag, attribute, value):
        '''
        return a named node from the XML Schema
        
        :param str tag: XML Schema tag (such as "complexType") to match
        :param str attribute: attribute name to match
        :param str value: attribute value to match
        '''
        root = cache.get_nxdl_xsd().getroot()
        xpath_str = 'xs:' + tag
        xpath_str += '[@' + attribute
        xpath_str += '="' + value + '"]'
        node_list = root.xpath(xpath_str, namespaces=self.ns)
        if len(node_list) != 1:
            msg = 'wrong number of ' + tag
            msg += ' nodes found: ' + str(len(node_list))
            raise ValueError(msg)
        return node_list[0]
    
    def copy_to(self, target):
        '''
        copy results into target object
        
        :param obj target: instance of Mixin, such as NXDL_Element
        '''
        for k, v in self.attrs.items():
            target.attrs[k] = v
        for k, v in self.children.items():
            target.children[k] = v

    def parse_attribute(self, node):
        ''' '''
        obj = Attribute(node)
        self.attrs[obj.name] = obj

    def parse_attributeGroup(self, node):
        ''' '''
        obj = NXDL_Type(node.attrib.get('ref'))
        obj.copy_to(self)

    def parse_complexContent(self, node):
        ''' '''
        for subnode in node:
            if subnode.tag.endswith('}extension'):
                ref = subnode.attrib.get('base')
                if ref not in ('nx:basicComponent'):
                    self.raise_error(subnode, 'unexpected base=', ref)
                obj = NXDL_Type(ref)
                obj.copy_to(self)

                # parse children of extension node
                for obj_node in subnode:
                    if obj_node.tag.endswith('}annotation'):
                        pass
                    elif obj_node.tag.endswith('}attribute'):
                        self.parse_attribute(obj_node)
                    elif obj_node.tag.endswith('}sequence'):
                        self.parse_sequence(obj_node)
                    else:
                        self.raise_error(obj_node, 'unexpected base=', obj_node.tag)

            else:
                self.raise_error(subnode, 'unexpected tag=', subnode.tag)

    def parse_group(self, node):
        ''' '''
        obj = NXDL_Type(node.attrib.get('ref'))
        obj.copy_to(self)


class NXDL_Root(Mixin):
    '''
    root of the nxdl.xsd file
    
    :param lxml.etree.Element xml_obj: XML element
    :param str obj_name: optional, default taken from ``xml_obj``
    :param dict ns_dict: optional, default taken from :data:`NAMESPACE_DICT`
    '''
    
    def __init__(self, xml_obj, obj_name=None, ns_dict=None):
        Mixin.__init__(self, xml_obj, obj_name=obj_name, ns_dict=ns_dict)
        self.attrs = {}
        self.children = {}

        element_type = xml_obj.attrib.get('type')
        if element_type is None:
            element_name = xml_obj.attrib.get('name')
            self.raise_error(xml_obj, 'no @type for element node: ', element_name)
        
        ref = strip_ns(element_type)
        type_node = self.get_named_node('complexType', 'name', ref)
        
        for node in type_node:
            if node.tag.endswith('}attribute'):
                obj = Attribute(node)
                self.attrs[obj.name] = obj
            elif node.tag.endswith('}attributeGroup'):
                self.parse_attributeGroup(node)
            elif node.tag.endswith('}sequence'):
                self.parse_sequence(node)
            elif node.tag.endswith('}annotation'):
                pass
            else:
                self.raise_error(node, 'unhandled tag=', node.tag)

    def parse_sequence(self, seq_node):
        '''
        parse the sequence used in the root element
        '''
        for node in seq_node:
            if node.tag.endswith('}element'):
                obj = NXDL_Element(node)
                self.children[obj.name] = obj
            elif node.tag.endswith('}group'):
                obj = NXDL_Type(node.attrib.get('ref'))
                obj.copy_to(self)
            else:
                msg = 'unhandled tag in ``definitionType``: '
                self.raise_error(node, msg, node.tag)


class Attribute(Mixin): 
    '''
    nx:attribute element
    
    :param lxml.etree.Element xml_obj: XML element
    :param str obj_name: optional, default taken from ``xml_obj``
    :param dict ns_dict: optional, default taken from :data:`NAMESPACE_DICT`
    '''
    
    def __init__(self, xml_obj, obj_name=None, ns_dict=None):
        Mixin.__init__(self, xml_obj, obj_name=obj_name, ns_dict=ns_dict)

        use = xml_obj.attrib.get('use', 'optional')
        self.required = use in ('required', )

        self.type = xml_obj.attrib.get('type', 'str')
        defalt = xml_obj.attrib.get('default')
        if self.type in ('nx:NX_BOOLEAN',):
            self.default_value = defalt.lower() in ('true', 'y', 1)
        else:
            self.default_value = defalt

        self.allowed_values = []
        xpath_str = 'xs:simpleType/xs:restriction/xs:enumeration'
        for node in xml_obj.xpath(xpath_str, namespaces=self.ns):
            v = node.attrib.get('value')
            if v is not None:
                self.allowed_values.append(v)

        self.patterns = []
        xpath_str = 'xs:simpleType/xs:restriction/xs:pattern'
        for node in xml_obj.xpath(xpath_str, namespaces=self.ns):
            v = node.attrib.get('value')
            if v is not None:
                self.patterns.append(v)


class NXDL_Element(Mixin): 
    '''
    an element
    
    :param lxml.etree.Element xml_obj: XML element
    :param str obj_name: optional, default taken from ``xml_obj``
    :param dict ns_dict: optional, default taken from :data:`NAMESPACE_DICT`
    
    :see: http://download.nexusformat.org/doc/html/nxdl.html
    :see: http://download.nexusformat.org/doc/html/nxdl_desc.html#nxdl-elements
    '''
    
    def __init__(self, xml_obj, obj_name=None, ns_dict=None):
        global __group_parsing_started__
        Mixin.__init__(self, xml_obj, obj_name=obj_name, ns_dict=ns_dict)
        self.children = {}
        self.attrs = {}

        # read & analyze theNXDL structural *type* referenced by *ref*
        ref = self.type = xml_obj.attrib.get('type')
        if ref is None:
            for node in xml_obj:
                if node.tag.endswith('}complexType'):
                    a = Attribute(node.find('xs:attribute', self.ns))
                    self.attrs[a.name] = a
                elif node.tag.endswith('}annotation'):
                    pass
                else:
                    self.raise_error(node, 'unhandled tag=', node.tag)
        else:
            # avoid known infinite recursion: group may contain group(s)
            ok_to_parse = True
            if xml_obj.attrib['name'] == 'group' and xml_obj.attrib['type'] == 'nx:groupType':
                if __group_parsing_started__:
                    ok_to_parse = False
                    # needs a special code to apply this rule
                    #     isinstance(obj, Recursion)
                    self.children['group'] = Recursion('group')
                __group_parsing_started__ = True
            if ok_to_parse:
                type_obj = NXDL_Type(ref)
                type_obj.copy_to(self)


class NXDL_Type(Mixin): 
    '''
    a named NXDL structure type (such as groupGroup)
    
    :param str ref: name of NXDL structure type (such as ``groupGroup``)
    :param str tag: XML Schema element tag, such as complexType (default=``*``)
    
    :see: http://download.nexusformat.org/doc/html/nxdl.html
    :see: http://download.nexusformat.org/doc/html/nxdl_desc.html#nxdl-data-types-internal
    '''
    
    def __init__(self, ref, tag = '*'):
        # Mixin.__init__(self, xml_obj)
        # do the Mixin.__init__ directly here
        self.ns = NAMESPACE_DICT

        xml_obj = self.get_named_node(tag, 'name', strip_ns(ref))
        self.name = xml_obj.attrib.get('name')
        
        self.attrs = {}
        self.children = {}

        for node in xml_obj:
            if isinstance(node, lxml.etree._Comment):
                pass
            elif node.tag.endswith('}annotation'):
                pass
            elif node.tag.endswith('}attribute'):
                self.parse_attribute(node)
            elif node.tag.endswith('}attributeGroup'):
                self.parse_attributeGroup(node)
            elif node.tag.endswith('}complexContent'):
                self.parse_complexContent(node)
            elif node.tag.endswith('}group'):
                self.parse_group(node)
            elif node.tag.endswith('}sequence'):
                self.parse_sequence(node)
            else:
                self.raise_error(node, 'unexpected tag=', node.tag)

    def parse_sequence(self, node):
        ''' '''
        for subnode in node:
            if subnode.tag.endswith('}element'):
                obj = NXDL_Element(subnode)
                self.children[obj.name] = obj
            elif subnode.tag.endswith('}group'):
                obj = NXDL_Element(subnode)
                self.children[obj.name] = obj
            elif subnode.tag.endswith('}any'):
                # do not process this one, only used for documentation
                pass
            else:
                self.raise_error(subnode, 'unexpected tag=', subnode.tag)


class Recursion(Mixin): 
    '''
    an element used in recursion, such as child group of group
    
    :param str obj_name: optional, default taken from ``xml_obj``
    '''
    
    def __init__(self, obj_name):
        Mixin.__init__(self, None, obj_name=obj_name, ns_dict=None)


def main():
    rules = NxdlRules()
    print rules


if __name__ == '__main__':  # remove for production
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
validate NeXus NXDL and HDF5 data files

These are considerata for the validation of NeXus HDF5 data files.
Compare these validation steps with rules and documentation
in the NeXus manual and the XML Schema files (``nxdl.xsd`` and ``nxdlTypes.xsd``).
Checkboxes indicate which steps have been implemented in code below.

* [x] make a list of all address nodes in the file to be evaluated
* [x] attributes are also in this list
* [x] use a structure to hold results for each node

.. rubric:: File

#. [x] verify attributes
#. [x] verify file level as group using NX_class = NXroot
#. [x] identify any objects at root level that are not in NXroot (which is OK)
#. [x] verify classpath exists: /NXentry/NXdata & @signal
#. [x] verify classpath exists: NXdata & @signal
#. [x] verify file has valid /NXentry/NXdata/signal_data
#. [x] verify every NXentry has NXdata/signal_data
#. [x] verify every NXdata has signal_data

    #. [x] version 1
    #. [x] version 2
    #. [x] version 3

.. rubric:: Groups

#. [x] compare name with pattern *validItemName*
#. [x] determine NX_class, if any
#. [x] verify NX_class in nxdl_dict
#. [ ] is name flexible?
#. [ ] What to do with NXDL symbol tables?
#. [ ] observe attributes: minOccurs maxOccurs
#. [ ] is deprecated?
#. [ ] special cases:

    #. [ ] NXentry
    #. [ ] NXsubentry
    #. [~] NXdata
    #. [x] NXcollection

#. [x] check for items defined by NX_class
#. [ ] check for items required by NX_class
#. [x] check for items not defined by NX_class
#. [ ] observe NXDL setting: ignoreExtraGroups
#. [ ] observe NXDL setting: ignoreExtraFields
#. [ ] observe NXDL setting: ignoreExtraAttributes
#. [x] validate any attributes
#. [x] validate any links
#. [x] validate any fields

.. rubric:: Links

#. [x] compare name with pattern *validItemName*
#. [ ] is name flexible?
#. [x] is target attribute defined?
#. [ ] verify target attribute with pattern *validTargetName*
#. [x] is target address absolute?
#. [x] does target address exist?
#. [x] construct NX classpath from target
#. [ ] compare NX classpath with NXDL specification

.. rubric:: Fields

#. [x] compare name with pattern
#. [ ] is name flexible?
#. [ ] observe attributes: minOccurs maxOccurs
#. [ ] is deprecated?
#. [x] is units attribute defined?
#. [ ] check units are consistent against NXDL
#. [ ] check data shape against NXDL
#. [ ] check data type against NXDL
#. [x] check for attributes defined by NXDL
#. [x] check AXISNAME_indices are each within signal data rank

.. rubric:: Attributes

#. [x] compare name with pattern
#. [ ] is deprecated?
#. [x] check data type against NXDL
#. [ ] check nxdl.xsd for how to handle these attributes regarding finding.WARN

    #. [ ] restricts
    #. [ ] ignoreExtraGroups
    #. [ ] ignoreExtraFields
    #. [ ] ignoreExtraAttributes

'''

import collections
import h5py
import lxml.etree
import numpy
import re

import cache
import finding
import h5structure
import nxdlstructure


# TODO: http://download.nexusformat.org/doc/html/search.html?q=warning&check_keywords=yes&area=default

__url__ = 'http://punx.readthedocs.org/en/latest/validate.html'

# for each NeXus data type, make a list of acceptable Python data types
# Is there a better way to define these?  Using nxdlTypes.xsd?
NXDL_DATA_TYPES = {
    'NX_CHAR': (str, unicode, numpy.string_, numpy.ndarray),
    'NX_INT':  (int, numpy.int, numpy.int8, numpy.int16, numpy.int32, numpy.int64),
    'NX_FLOAT':  (float, ),
    'NX_BINARY': (None, ),     # FIXME:
    'NX_BOOLEAN': (None, ),     # FIXME:
}
NXDL_DATA_TYPES['NX_UINT']   = NXDL_DATA_TYPES['NX_INT']
NXDL_DATA_TYPES['NX_POSINT'] = NXDL_DATA_TYPES['NX_INT']
NXDL_DATA_TYPES['NX_NUMBER'] = NXDL_DATA_TYPES['NX_INT'] + NXDL_DATA_TYPES['NX_FLOAT']
NXDL_DATA_TYPES['ISO8601']   = NXDL_DATA_TYPES['NX_CHAR']


def validate_xml(xml_file_name):
    '''
    validate an NXDL XML file against an XML Schema file

    :param str xml_file_name: name of XML file
    '''
    xml_tree = lxml.etree.parse(xml_file_name)
    xsd = cache.get_XML_Schema()
    return xsd.assertValid(xml_tree)


class NxdlPattern(object):
    '''
    common regular expression pattern for validation
    
    :param obj parent: instance of :class:`Data_File_Validator`
    :param str pname: pattern identifying name
    :param str xpath_str: XPath search string, expect list of length = 1
    '''
    
    def __init__(self, parent, pname, xpath_str):
        self.name = pname
        self.xpath_str = xpath_str

        r = parent.nxdl_xsd.xpath(xpath_str, namespaces=parent.ns)

        if r is None or len(r) != 1:
            msg = 'could not read *' + pname + '* from *nxdl.xsd*'
            raise ValueError(msg)

        self.regexp_pattern_str = r[0].attrib.get('value', None)
        self.re_obj = re.compile('^' + self.regexp_pattern_str + '$')
    
    def match(self, text):
        '''regular expression search'''
        return self.re_obj.match(text)


class CustomNxdlPattern(NxdlPattern):
    '''
    custom regular expression pattern for validation
    
    :param obj parent: instance of :class:`Data_File_Validator`
    :param str pname: pattern identifying name
    :param str regexp_pattern_str: regular expression to match
    '''
    
    def __init__(self, parent, pname, regexp_pattern_str):
        self.name = pname
        self.xpath_str = None

        self.regexp_pattern_str = regexp_pattern_str
        self.re_obj = re.compile('^' + self.regexp_pattern_str + '$')
    
    def match(self, text):
        '''regular expression search'''
        return self.re_obj.match(text)


class Data_File_Validator(object):
    '''
    manage the validation of a NeXus HDF5 data file
    '''
    
    def __init__(self, fname):
        self.fname = fname
        self.findings = []      # list of Finding() instances
        self.addresses = collections.OrderedDict()     # dictionary of all HDF5 address nodes in the data file

        # open the NXDL rules files
        #cache.update_NXDL_Cache()        # let the user control when to update

        self.ns = cache.NX_DICT
        self.nxdl_xsd = cache.get_nxdl_xsd()
        nxdlTypes_xsd_file = cache.abs_NXDL_filename(cache.NXDL_TYPES_SCHEMA_FILE)
        self.nxdlTypes_xsd = lxml.etree.parse(nxdlTypes_xsd_file)

        self.nxdl_dict = nxdlstructure.get_NXDL_specifications()
        self.h5 = h5py.File(fname, 'r')
        self._init_patterns()
    
    def _init_patterns(self):
        self.patterns = {}
        for item in ('validItemName', 'validNXClassName', 
                     'validTargetName'):
            xps = '//*[@name="' # XPath String query
            xps += item
            xps += '"]/xs:restriction/xs:pattern'
            self.patterns[item] = NxdlPattern(self, item, xps)

        # strict match: [a-z_][a-z\d_]*
        # flexible match: [A-Za-z_][\w_]*  but gets finding.WARN per manual
        # advisory changed to finding.NOTE
        p = CustomNxdlPattern(self, 'validItemName-strict', r'[a-z_][a-z0-9_]*')
        self.patterns[p.name] = p
    
    def validate_default_plot(self):
        '''
        check that data file defines the default plottable data
        
        :see: http://download.nexusformat.org/doc/html/datarules.html#find-the-plottable-data
        '''
        # - - - - - - - - - -
        def find_NXdata_in_group(group):
            if h5structure.isNeXusGroup(group, 'NXdata'):
                nxdata_dict[group.name] = group
            for h5_child in sorted(group):
                h5_obj = group.get(h5_child)
                if h5structure.isHdf5Group(h5_obj):
                    find_NXdata_in_group(h5_obj)
        # - - - - - - - - - -
        def version1(group):
            '''plottable data version 1
            
            :see: http://download.nexusformat.org/doc/html/datarules.html#design-findplottable-bydimnumber
            '''
            base_title = 'default plot v1'
            dimension_scales = []
            signal_count = 0
            for field_name in group:
                field = group.get(field_name)
                signal = self.get_hdf5_attribute(field, 'signal')
                primary = self.get_hdf5_attribute(field, 'primary')
                if signal is not None:
                    title = base_title + ' @signal attribute'
                    msg = {True: 'value: ' + str(signal), False: 'does not exist'}
                    self.new_finding(title, field.name + '@signal', finding.OK, msg)
                    try:
                        int_signal = int(signal)
                        if int_signal == 1:
                            signal_count += 1
                            title = base_title + ' @signal value checked'
                            self.new_finding(title, field.name + '@signal', finding.OK, 'default data')
                    except ValueError:
                        title = base_title + ' @signal value checked'
                        m = '@signal="' + str(signal) + '" not interpreted as integer'
                        self.new_finding(title, field.name + '@signal', finding.ERROR, m)
                elif primary is not None:
                    title = base_title + ' @primary attribute'
                    msg = {True: 'value: ' + str(primary), False: 'does not exist'}
                    self.new_finding(title, field.name + '@primary', finding.OK, 'value: ' + msg)
                    try:
                        int_primary = int(primary)
                        if int_primary == 1:
                            title = base_title + ' @primary value checked'
                            m = 'possible dimension scale'
                            self.new_finding(title, field.name + '@primary', finding.OK, m)
                            # TODO: verify that shape is 1-D
                            m = 'verify the shape is 1-D'
                            self.new_finding(title, field.name + '@primary', finding.TODO, m)
                            dimension_scales.append(field_name)
                    except ValueError:
                        title = base_title + ' @primary value checked'
                        m = '@primary="' + str(signal) + '" not interpreted as integer'
                        self.new_finding(title, field.name + '@primary', finding.ERROR, m)
                title = 'field @axis attribute'
                self.new_finding(title, field.name + '@axis', finding.TODO, 'not checked')

            if signal_count == 0:
                m = 'no field marked with @signal="1"'
                self.new_finding(base_title, field.name, finding.ERROR, m)
            elif signal_count > 1:
                m = 'too many fields marked with @signal="1"'
                self.new_finding(base_title, field.name, finding.ERROR, m)
            return signal_count == 1
        # - - - - - - - - - -
        def version2(group):
            '''plottable data version 2
            
            :see: http://download.nexusformat.org/doc/html/datarules.html#design-findplottable-byname
            '''
            base_title = 'default plot v2'
            signal_count = 0
            for field_name in group:
                field = group.get(field_name)
                signal = self.get_hdf5_attribute(field, 'signal')
                t = signal is not None
                if t:
                    title = base_title + ' @signal attribute'
                    msg = {True: 'value: ' + str(signal), False: 'does not exist'}[t]
                    self.new_finding(title, field.name + '@signal', finding.OK, 'value: ' + msg)
                    try:
                        int_signal = int(signal)
                        if int_signal == 1:
                            signal_count += 1
                            title = base_title + ' @signal value checked'
                            self.new_finding(title, field.name + '@signal', finding.OK, 'default data')
                    except ValueError:
                        title = base_title + ' @signal value checked'
                        m = '@signal="' + str(signal) + '" not interpreted as integer'
                        self.new_finding(title, field.name + '@signal', finding.ERROR, m)
                title = base_title + ' @axes attribute'
                # axes="axis1:xis2:xis3"
                # delimiter could be either ":" or ","
                self.new_finding(title, field.name + '@axes', finding.TODO, 'not checked')
                
            return signal_count == 1
        # - - - - - - - - - -
        def version3(group):
            '''plottable data version 3
            
            :see: http://download.nexusformat.org/doc/html/datarules.html#design-findplottable-niac2014
            '''
            base_title = 'default plot v3'
            
            signal = self.get_hdf5_attribute(group, 'signal')
            t = signal is not None
            f = {True: finding.OK, False: finding.NOTE}[t]
            title = base_title + ' @signal attribute'
            msg = {True: 'exists', False: 'does not exist'}[t]
            self.new_finding(title, group.name, f, msg)
            if not t:
                return False

            t = signal in group
            f = finding.TF_RESULT[t]
            title = base_title + ' @signal value'
            msg = {True: 'value: ' + signal, False: 'does not exist'}[t]
            self.new_finding(title, group.name, f, msg)
            if not t:
                return False
            signal_obj = group.get(signal)
            signal_rank = len(signal_obj.shape)

            axes = self.get_hdf5_attribute(group, 'axes')
            t = axes is not None
            f = {True: finding.OK, False: finding.NOTE}[t]
            title = base_title + ' @axes attribute'
            msg = {True: 'exists', False: 'does not exist'}[t]
            self.new_finding(title, group.name, f, msg)
            if axes is not None:
                for a in axes.split(','):
                    t = a in group and a != signal
                    f = finding.TF_RESULT[t]
                    title = base_title + ' @axes value'
                    msg = {True: 'value: ' + a, False: 'does not exist'}[t]
                    self.new_finding(title, group.name, f, 'value: ' + a)

                    if t:
                        # AXISNAME_indices
                        axis_index = self.get_hdf5_attribute(group, a + '_indices')
                        t = axis_index is not None
                        f = {True: finding.OK, False: finding.NOTE}[t]
                        gn = group.name + '@' + a + '_indices'
                        title = base_title + ' @AXISNAME_indices attribute'
                        msg = {True: 'value: ' + str(axis_index), False: 'does not exist'}[t]
                        self.new_finding(title, gn, f, msg)

                        if t:
                            # value of AXISNAME_indices is within the rank of the signal data
                            title = base_title + ' @AXISNAME_indices value'
                            t = 0 <= axis_index < signal_rank
                            f = finding.TF_RESULT[t]
                            msg = 'rank test: 0 <= ' + str(axis_index) + ' < ' + str(signal_rank)
                            self.new_finding(title, gn, f, msg)

            self.new_finding('default plot v3 dimension scales', group.name, finding.TODO, 'unchecked')
            return True
        # - - - - - - - - - -
        def niac2015_default_path(group, subgroup_class):
            group_default = self.get_hdf5_attribute(group, 'default')
            title = '@default attribute'
            t = group_default is not None
            f = {True: finding.OK, False: finding.NOTE}[t]
            m = {True: 'exists', False: 'does not exist'}[t]
            gn = group.name or '/'
            self.new_finding(title, gn+'@default', f, m)

            if t:
                title = '@default value'
                t = group_default in group
                f = finding.TF_RESULT[t]
                m = 'HDF5 item ' + {True: 'exists', False: 'does not exist'}[t]
                self.new_finding(title, gn+'@default', f, m)
                
                if t:
                    title = '@default value test'
                    subgroup = group[group_default]
                    t = h5structure.isNeXusGroup(subgroup, subgroup_class)
                    m = 'HDF5 item is ' + subgroup_class
                    self.new_finding(title, gn+'@default', f, m)
            return group_default
        # - - - - - - - - - -

        # identify the NXdata groups to check, do not treat links differently
        nxdata_dict = collections.OrderedDict()
        find_NXdata_in_group(self.h5)

        title = 'NXdata default plottable data'
        valid_nxdata_dict = collections.OrderedDict()
        for h5_addr, nxdata_group in nxdata_dict.items():
            if version3(nxdata_group):
                self.new_finding(title, h5_addr, finding.OK, 'version 3')
                valid_nxdata_dict[h5_addr] = nxdata_group
            elif version2(nxdata_group):
                self.new_finding(title, h5_addr, finding.OK, 'version 2')
                valid_nxdata_dict[h5_addr] = nxdata_group
            elif version1(nxdata_group):
                self.new_finding(title, h5_addr, finding.OK, 'version 1')
                valid_nxdata_dict[h5_addr] = nxdata_group
            else:
                m = 'no default plottable data indicated'
                self.new_finding(title, h5_addr, finding.ERROR, m)

        # search each NXentry for valid NXdata
        for h5_addr in sorted(self.h5):
            nxentry = self.h5.get(h5_addr)
            if h5structure.isNeXusGroup(nxentry, 'NXentry'):
                count = 0
                for item in sorted(nxentry):
                    nxdata = nxentry.get(item)
                    if h5structure.isNeXusGroup(nxdata, 'NXdata'):
                        if nxdata.name in valid_nxdata_dict.keys():
                            count += 1
                t = count > 0
                f = finding.TF_RESULT[t]
                title = '/NXentry/NXdata/signal_data'
                m = 'basic NeXus requirement: default plot'
                m += {True: '', False: ' not'}[t] + ' described'
                self.new_finding(title, nxentry.name, f, m)

        title = '/NXroot@default/NXentry@default/NXdata'
        nxentry = niac2015_default_path(self.h5, 'NXentry')
        if nxentry is not None:
            nxdata = niac2015_default_path(self.h5[nxentry], 'NXdata')
            if nxdata is not None:
                m = 'absolute path to default plot'
                self.new_finding(title, self.h5[nxentry][nxdata].name, finding.OK, m)
    
    def review_with_NXDL(self, group, nx_class):
        '''
        review *group* with the NXDL specification for *nx_class*
        
        :param obj group: instance of h5py.Group or h5py.Dataset
        :param str nx_class: name of NeXus NXDL specification
        '''
        #self.new_finding('-'*10, group.name, finding.COMMENT, 'review_with_NXDL start' + '-'*10)
        
        if not h5structure.isHdf5File(group):
            nxc = self.get_hdf5_attribute(group, 'NX_class', '<undefined>')
            f = finding.TF_RESULT[nxc in self.nxdl_dict]
            msg = nxc  + ' known?'
            self.new_finding('NX_class known', group.name, f, 'value: ' + nxc)
        
        if nx_class in self.nxdl_dict:
            self.new_finding('NXDL known', group.name, finding.OK, nx_class)
        else:
            msg = 'expected NX_class unknown: ' + nx_class
            self.new_finding('NXDL known', group.name, finding.OK, msg)
            #self.new_finding('-'*10, group.name, finding.COMMENT, 'review_with_NXDL bailout' + '-'*10)
            return

        if group.attrs.get('NX_class', '') in ('NXcollection',):
            msg = 'NXcollection content is not validated'
            self.new_finding('NXcollection', group.name, finding.OK, msg)
            return

        self.validate_attributes(group, nx_class)
        
        nxdl_spec = self.nxdl_dict[nx_class]
        nxdl_subgroups = nxdl_spec.getSubGroup_NX_class_list()
        for h5_child in sorted(group):
            h5_obj = group.get(h5_child)
            if h5structure.isHdf5Group(h5_obj):
                h5_sub_nx_class = h5_obj.attrs.get('NX_class', None)
                s = h5_sub_nx_class in nxdl_subgroups
                f = {True: finding.OK, False: finding.NOTE}[s]
                c = h5_sub_nx_class
                if f == finding.NOTE:
                    s = h5_sub_nx_class in self.nxdl_dict
                    f = {True: finding.NOTE, False: finding.ERROR}[s]
                    c += ': not known '
                    c += {True: 'subgroup of ', False: 'NeXus class'}[s]
                    c += ' ' + nx_class
                self.new_finding(nx_class + ' subgroup', h5_obj.name, f, c)
            if h5structure.isHdf5Dataset(h5_obj):
                known_name = h5_obj.name in nxdl_spec.fields
                c = 'strict comparison'
                if not known_name:
                    # TODO: check here if name is flexible : How to do this?
                    known_name = nx_class in ('NXdata', 'NXdetector')
                    c = 'acceptable'
                f = {True: finding.OK, False:finding.NOTE}[known_name]
                self.new_finding(nx_class + ' defined field', h5_obj.name, f, c)
                msg = 'need to complete check with NXDL'
                self.new_finding(nx_class + ' field', h5_obj.name, finding.TODO, msg)

        # self.new_finding('-'*10, group.name, finding.COMMENT, 'review_with_NXDL end' + '-'*10)

    def validate(self):
        '''
        start the validation process from the file root
        '''
        self.new_finding('validate all names', '/', finding.COMMENT, '-'*10)
        self.collect_names(self.h5)

        # HDF5 group attributes
        for item in sorted(self.h5.attrs.keys()):
            aname = self.h5.name + '@' + item
            self.new_address(aname)
        
        # this may be useful for validating rule for default plot, for example
        # /NXentry/NXdata/<any>@signal
        # /NXentry/NXdata@signal
        self.classpath_dict = collections.OrderedDict()
        for k, v in self.addresses.items():
            self.classpath_dict[k] = v

        # self.new_finding('-'*10, '/', finding.COMMENT, 'NXroot checkup start' + '='*10)
        self.new_finding('validate all objects', '/', finding.COMMENT, '-'*10)
        title = 'hdf5 file root object'
        for item in sorted(self.h5):
            obj = self.h5.get(item)
            if h5structure.isNeXusLink(obj):
                self.validate_link(obj, self.h5)
            elif h5structure.isHdf5Group(obj):
                self.validate_group(obj, 'NXentry')
            elif h5structure.isHdf5Dataset(obj):
                self.validate_dataset(obj, self.h5)
            else:
                self.new_finding(title, obj.name, finding.NOTE, 'not a NeXus item')
        # self.new_finding('-'*10, '/', finding.COMMENT, 'NXroot checkup end' + '='*10)
        
        # review file with the relevant NXDL specification: NXroot
        # review of other classes as specified in the data file is called in validate_group()
        self.new_finding('review file root with NXroot', '/', finding.COMMENT, '-'*10)
        self.review_with_NXDL(self.h5, 'NXroot')

        self.new_finding('check for default plot', '/', finding.COMMENT, '-'*10)
        self.validate_default_plot()

    def validate_group(self, group, nxdl_classname):
        '''
        check group against the specification of nxdl_classname
        
        :param obj group: instance of h5py.Group
        :param str nxdl_classname: name of NXDL class this group should match
        '''
        nx_class = self.get_hdf5_attribute(group, 'NX_class')
        if nx_class is None:
            comment = 'hdf5 group has no `NX_class` attribute'
            self.new_finding('HDF5 group', group.name, finding.NOTE, comment)
        else:
            self.review_with_NXDL(group, nx_class)
        
        self.validate_attributes(group, nx_class)
        
        # get a list of the NXDL subgroups defined in this group
        nxdl_class_obj = self.nxdl_dict[nxdl_classname]
        defined_nxdl_list = nxdl_class_obj.getSubGroup_NX_class_list()
        
        # HDF5 group children
        for item in sorted(group):
            obj = group.get(item)
            if h5structure.isNeXusLink(obj):
                # pull these out BEFORE groups & fields
                self.validate_link(obj, group)
            elif h5structure.isHdf5Group(obj):
                obj_nx_class = self.get_hdf5_attribute(obj, 'NX_class')
                if obj_nx_class in defined_nxdl_list:
                    self.validate_group(obj, obj_nx_class)
                else:
                    self.new_finding('defined', obj.name, finding.NOTE, 'not defined in ' + nxdl_classname)
            elif h5structure.isHdf5Dataset(obj):
                self.validate_dataset(obj, group)
            else:
                self.new_finding('dataset', obj.name, finding.TODO, finding.TODO.description)

    
    def validate_dataset(self, dataset, group):
        '''
        check dataset against the specification of group NXDL specification
        
        :param obj dataset: instance of h5py.Dataset
        :param obj group: instance of h5py.Group
        '''
#         ds_name = dataset.name.split('/')[-1]
        if h5structure.isHdf5File(group):
            nx_class = 'NXroot'
        else:
            nx_class = self.get_hdf5_attribute(group, 'NX_class')
        nxdl_class_obj = self.nxdl_dict.get(nx_class, None)
        if nxdl_class_obj is None:
            self.new_finding('unknown NX_class', dataset.name, finding.ERROR, 'found: ' + nx_class)
        # if the data type is NX_NUMBER, is @units defined and has *some* value?
        if dataset.dtype in NXDL_DATA_TYPES['NX_NUMBER']:
            title = 'field data type'
            msg = 'data type: ' + str(dataset.dtype)
            self.new_finding(title, dataset.name, finding.OK, msg)
            
            title = 'field units attribute'
            units = self.get_hdf5_attribute(dataset, 'units')
            t = units is not None
            f = {True: finding.OK, False: finding.NOTE}[t]
            msg = {True: 'exists', False: 'does not exist'}[t]
            self.new_finding(title, dataset.name + '@units', f, msg)
            if t:
                t = len(units) > 0
                f = {True: finding.OK, False: finding.NOTE}[t]
                msg = {True: 'value: ' + units, False: 'has no value'}[t]
                self.new_finding(title, dataset.name + '@units', f, msg)

        self.validate_attributes(dataset, nx_class)

    def validate_link(self, link, group):
        '''
        check link against the specification of nxdl_classname
        
        :param obj link: instance of h5py.Group or h5py.Dataset
        :param obj group: instance of h5py.Group, needed to check against NXDL
        '''
        target = link.attrs.get('target', None)
        if target is not None:
            target_exists = target in self.h5
            f = finding.TF_RESULT[target_exists]
            msg = {True: target, False: 'does not exist'}[target_exists]
            self.new_finding('link target', link.name, f, msg)
            # not necessary: match target nexus classpath and match with NXDL
        else:
            self.new_finding('link', link.name, finding.ERROR, 'no target')

    def validate_attributes(self, h5_obj, nxdl_class):
        '''
        check attributes of obj against the specification of nxdl_classname
        
        :param obj obj: instance of h5py object with attributes
        :param str nxdl_class: NXDL class name
        '''
        if nxdl_class not in self.nxdl_dict:
            status = finding.ERROR
            msg = 'unknown: ' + nxdl_class
            self.new_finding('NXDL NX_class', h5_obj, status, msg)
            return

        if nxdl_class in ('NXcollection',):
            msg = 'NXcollection content is not validated'
            self.new_finding('NXcollection', h5_obj.name, finding.OK, msg)
            return

        nxdl_class_obj = self.nxdl_dict[nxdl_class]
        if h5structure.isHdf5Group(h5_obj):
            title = nxdl_class + ' attribute'
        elif h5structure.isHdf5Dataset(h5_obj):
            title = 'field attribute'
        else:
            raise ValueError('unknown object: ' + h5_obj.name)
        tf_result = {True: finding.OK, False: finding.UNUSED}

        # get list of all possible attributes from data file and NXDL spec
        h5_attrs = h5_obj.attrs.keys() + nxdl_class_obj.attrs.keys()
        h5_attrs = map(str, {k:None for k in h5_attrs}.keys())      # remove extras
        if 'NX_class' in h5_attrs:
            h5_attrs.remove('NX_class')

        attributes_reviewed_elsewhere = [
            'NX_class AXISNAME_indices signal axes axis primary units'.split()
        ]
        for k in sorted(h5_attrs):
            aname = h5_obj.name + '@' + k
            if k in attributes_reviewed_elsewhere:
                msg = 'reviewed elsewhere'
                self.new_finding(title, aname, finding.COMMENT, msg)
                continue
            data_type_checked = False
            if k in nxdl_class_obj.attrs:
                msg = 'defined in ' + nxdl_class
                status = tf_result[k in h5_obj.attrs]

                if k in h5_obj.attrs:                # check expected NXDL data type
                    data_type_checked = True
                    obj_attr = h5_obj.attrs[k]
                    nxdl_attr = nxdl_class_obj.attrs[k]
                    nx_type = nxdl_attr.nx_type
                    data_type_ok = nx_type in NXDL_DATA_TYPES and type(obj_attr) in NXDL_DATA_TYPES[nx_type]
                if status not in (finding.UNUSED,):
                    self.new_finding(title, aname, status, msg)
#             else:
#                 status = finding.NOTE
#                 msg = 'not defined in ' + nxdl_class
#             # TODO: need to learn *minOccurs* from NXDL
#             msg += ' (optional)'
            if data_type_checked:
                msg = str(type(obj_attr)) + ' : ' + nx_type
                self.new_finding('NXDL NX_type', aname, finding.TF_RESULT[data_type_ok], msg)

    def collect_names(self, h5_object):
        '''
        get the fullname of this object and any of its children
        
        also, check this name with the NeXus 
        *validItemName* regular expression
        '''
        self.new_address(h5_object.name)
        if not h5structure.isHdf5File(h5_object):
            self.validate_item_name(h5_object.name)
        for item in sorted(h5_object.attrs.keys()):
            aname = h5_object.name + '@' + item
            self.new_address(aname)
            self.validate_item_name(aname)
        
        if h5structure.isHdf5Group(h5_object):
            for item in sorted(h5_object):
                obj = h5_object.get(item)
                if h5structure.isNeXusLink(obj):
                    # pull these out BEFORE groups & fields
                    self.new_address(obj.name)
                    self.validate_item_name(obj.name)
                else:
                    # anything else
                    self.collect_names(obj)

    def validate_item_name(self, h5_addr):
        '''
        validate *h5_addr* using *validItemName* regular expression
        
        This is used for the names of groups, fields, links, and attributes.
        
        :param str h5_addr: full HDF5 address of item, for reference only,
            for attributes, use an @ symbol, such as these examples:
            
            =============================    ============
            *h5_addr*                        *short_name*
            =============================    ============
            ``/entry/user``                  ``user``
            ``/entry/data01/data``           ``data``
            ``/entry/data01/data@signal``    ``signal``
            =============================    ============

        This method will separate out the last part of the name for validation.
        
        :see: http://download.nexusformat.org/doc/html/datarules.html?highlight=regular%20expression
        '''
        key_relaxed = 'validItemName'
        key_strict = 'validItemName-strict'

        # h5_addr = obj.name
        short_name = h5_addr.split('/')[-1].split('@')[-1]
        
        # strict match: [a-z_][a-z\d_]*
        # flexible match: [A-Za-z_][\w_]*  but gets finding.WARN per manual

        p = self.patterns[key_strict]
        m = p.match(short_name)
        if m is not None and m.string == short_name:
            name_ok = finding.OK
            adjective = 'strict'
            key = key_strict
        else:
            p = self.patterns[key_relaxed]
            m = p.match(short_name)
            t = m is not None and m.string == short_name
            # opinion: finding.WARN is too harsh
            #     NXroot defines attributes that produce such warnings
            #     capital letters in names are generally acceptable, only a few complain
            name_ok = {True: finding.NOTE, False: finding.ERROR}[t]
            adjective = 'relaxed'
            key = key_relaxed

        self.new_finding(key, h5_addr, name_ok, adjective +' re: ' + p.regexp_pattern_str)

    def new_finding(self, test_name, h5_address, status, comment):
        '''
        accumulate a list of findings
        '''
        addr = str(h5_address)
        f = finding.Finding(test_name, addr, status, comment)
        self.findings.append(f)
        if addr in self.addresses:
            self.addresses[addr].findings.append(f)

    def new_address(self, h5_address, *args, **kwargs):
        '''
        accumulate a dictionary of HDF5 object addresses
        '''
        self.addresses[h5_address] = finding.CheckupResults(h5_address)
        self.reconstruct_classpath(h5_address)

    def reconstruct_classpath(self, h5_address, *args, **kwargs):
        '''
        build the classpath from the h5_address
        '''
        path = h5_address.lstrip('/').split('@')[0]
        if len(path) == 0:
            return

        # reconstruct the NeXus classpath
        cp = ''     # classpath to be built
        hp = ''     # HDF5 address to be built
        for item in path.split('/'):
            hp += '/' + item
            if hp in self.h5:
                if h5structure.isHdf5Dataset(self.h5[hp]):
                    cp += '/' + item
                else:
                    obj = self.h5[hp]
                    nx_class = self.get_hdf5_attribute(self.h5[hp], 'NX_class', '-')
                    cp += '/' + nx_class
        if '@' in h5_address:
            cp += '@' + h5_address.split('@')[-1]
        
        if h5_address in self.addresses:
            self.addresses[h5_address].classpath = cp


    def get_hdf5_attribute(self, obj, attribute, default=None):
        '''
        HDF5 attribute strings might be coded in several ways
        '''
        a = obj.attrs.get(attribute, default)
        if isinstance(a, numpy.ndarray):
            gname = obj.name + '@' + attribute
            msg = '[variable length string]: ' + str(a)
            self.new_finding('attribute data type', gname, finding.NOTE, msg)
            a = a[0]
        return a
    
    def report_findings(self, statuses=()):
        '''
        make a table of the validation findings
        
        :param statuses: List (or tuple) of finding statuses to be shown.
            Several lists have been pre-defined for convenience:

            :data:`finding.SHOW_ALL`        ``(OK, NOTE, WARN, ERROR, TODO, UNUSED, COMMENT)``
            :data:`finding.SHOW_NOT_OK`     ``(ERROR, WARN)``
            :data:`finding.SHOW_ERRORS`     ``(WARN, ERROR, TODO, UNUSED)``
            
            See :mod:`finding` for details.

        '''
        import pyRestTable

        t = pyRestTable.Table()
        t.labels = 'address validation status comment(s)'.split()
        for f in self.findings:
            if f.status in statuses:
                t.rows.append((f.h5_address, f.test_name, f.status, f.comment))
        return t.reST()
    
    def report_findings_summary(self):
        '''
        make a summary table of the validation findings (count how many of each status)
        '''
        import pyRestTable

        # count each category
        summary = collections.OrderedDict()
        for k in finding.VALID_STATUS_LIST:
            summary[str(k.key)] = 0
        xref = {str(k): k for k in finding.VALID_STATUS_LIST}
        for f in self.findings:
            summary[str(f.status)] += 1

        t = pyRestTable.Table()
        t.labels = 'status count description'.split()
        for k, v in summary.items():
                t.rows.append((k, v, xref[k].description))
        t.rows.append(('--', '--', '--'))
        t.rows.append(('TOTAL', len(self.findings), '--'))
        return t.reST()
    
    def report_classpath(self):
        import pyRestTable
        t = pyRestTable.Table()
        t.labels = 'HDF5-address  NeXus-classpath'.split()
        for k, v in self.classpath_dict.items():
            t.rows.append((k, v))
        return t.reST()


if __name__ == '__main__':
    print "Start this module using:  python main.py validate ..."
    exit(0)




#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
document each item during validation
'''


class ValidationResultStatus(object):
    '''
    summary result of a Finding
    
    :param str key: short name
    :param str color: suggested color for GUI
    :param str description: one-line summary
    '''
    
    def __init__(self, key, color, description):
        self.key = key
        self.color = color
        self.description = description
    
    def __str__(self, *args, **kwargs):
        return self.key


OK      = ValidationResultStatus('OK',      'green',     'meets NeXus specification')
NOTE    = ValidationResultStatus('NOTE',    'palegreen', 'does not meet NeXus specification, but acceptable')
WARN    = ValidationResultStatus('WARN',    'yellow',    'does not meet NeXus specification, not generally acceptable')
ERROR   = ValidationResultStatus('ERROR',   'red',       'violates NeXus specification')
TODO    = ValidationResultStatus('TODO',    'blue',      'validation not implemented yet')
UNUSED  = ValidationResultStatus('UNUSED',  'grey',      'optional NeXus item not used in data file')
COMMENT = ValidationResultStatus('COMMENT', 'grey',      'comment from the punx source code')

VALID_STATUS_LIST = (OK, NOTE, WARN, ERROR, TODO, UNUSED, COMMENT)
TF_RESULT = {True: OK, False: ERROR}

SHOW_ALL = VALID_STATUS_LIST
SHOW_ERRORS = (ERROR, WARN)
SHOW_NOT_OK = (WARN, ERROR, TODO, UNUSED)


class Finding(object):
    '''
    a single observation noticed while validating
    
    :param str test_name: one-word description of the test
    :param str h5_address: address of h5py item
    :param int status: one of: OK NOTE WARNING ERROR TODO
    :param str comment: description
    '''
    
    def __init__(self, test_name, h5_address, status, comment):
        if status not in VALID_STATUS_LIST:
            msg = 'unknown status value: ' + status
            raise ValueError(msg)
        self.test_name = str(test_name)
        self.h5_address = h5_address
        self.status = status
        self.comment = comment
    
    def __str__(self, *args, **kwargs):
        s = self.status
        s += ' ' + self.h5_address 
        s += ': ' + self.test_name
        s += ': ' + self.comment
        return s


class CheckupResults(object):
    '''
    various checkups for a single hdf5 address (absolute path)
    
    :param str h5_address: address of h5py item
    '''
    
    def __init__(self, h5_address):
        self.h5_address = h5_address
        self.findings = []      # keep list of all findings for this address
        self.classpath = ''
    
    def __str__(self, *args, **kwargs):
        return self.h5_address



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
Load and/or document the structure of a NeXus NXDL class specification

* :class:`NXDL_definition`: the structure
* define a text renderer method for that class
'''

__url__ = 'http://punx.readthedocs.org/en/latest/nxdlstructure.html'

# testing:  see file dev_nxdl2rst.py


import collections
import lxml.etree
import os

import __init__
import cache
import nxdl_rules
import validate


PROGRAM_NAME = 'nxdlstructure'
INDENTATION_UNIT = '  '
listing_category = None

__singleton_nxdl_rules__ = None


def get_nxdl_rules():
    '''
    parse and cache the XML Schema file (nxdlTypes.xsd) as an XML document only once
    '''
    global __singleton_nxdl_rules__

    if __singleton_nxdl_rules__ is None:
        __singleton_nxdl_rules__ = nxdl_rules.NxdlRules()

    return __singleton_nxdl_rules__


def validate_NXDL(nxdl_file_name):
    '''
    Validate a NeXus NXDL file
    '''
    validate.validate_xml(nxdl_file_name)


class NXDL_mixin(object):
    '''
    common components available to all subclasses
    
    :param obj node: XML object
    '''
    element = 'mixin - must override in subclass'
    
    def __init__(self, node):
        node_name = node.get('name')
        if node_name is not None:
            self.name = node.get('name')
        self.ns = nxdl_rules.NAMESPACE_DICT
        
        # TODO: apply nxdl_rules.NxdlRules() as defaults when creating NXDL_mixin subclasses below
        self.nxdl_rules = get_nxdl_rules()
    
    def get_NX_type(self, node):
        '''
        '''
        return node.get('type', 'NX_CHAR')
    
    def get_NX_units(self, node):
        '''
        '''
        return node.get('units', '')

    def __str__(self):
        '''
        '''
        return '-tba-'
    
    def get_element_data(self, node, category):
        '''
        '''
        if self.element == 'definition':
            rules = self.nxdl_rules.nxdl
        else:
            rules = self.nxdl_rules.nxdl.children[self.element]
        self.attrs = {}
        self.fields = {}
        self.groups = {}
        self.links = {}

        for subnode in node.xpath('nx:attribute', namespaces=self.ns):
            obj = NX_attribute(subnode)
            self.add_object(self.attrs, obj)

        for subnode in node.xpath('nx:field', namespaces=self.ns):
            obj = NX_field(subnode, category)
            self.add_object(self.fields, obj)

        for subnode in node.xpath('nx:group', namespaces=self.ns):
            obj = NX_group(subnode, category)
            self.add_object(self.groups, obj)
        
        for subnode in node.xpath('nx:link', namespaces=self.ns):
            obj = NX_link(subnode, category)
            self.add_object(self.links, obj)
        
        pass
    
    def add_object(self, db, obj):
        '''
        '''
        name = obj.name
        if name in db:
            name += '_1'
        db[name] = obj

    def render_group(self, group):
        '''
        '''
        indentation = ' '*2
        t = []
        for _k, v in sorted(group.attrs.items()):
            t.append(str(v))
        for _k, v in sorted(group.fields.items()):
            t.append(str(v))
        for _k, v in sorted(group.links.items()):
            t.append(str(v))
        for _k, v in sorted(group.groups.items()):
            t.append(str(v))
            for line in group.render_group(v).splitlines():
                t.append(indentation + line)
        return '\n'.join(t)


class NXDL_definition(NXDL_mixin):
    '''
    Contains the complete structure of a NXDL definition, without documentation
    
    :param str nxdl_file: name of file with NXDL definition (ends with ``.nxdl.xml``)
    '''
    element = 'definition'
    
    def __init__(self, nxdl_file):
        self.nxdl_file_name = nxdl_file
        if not os.path.exists(nxdl_file):
            raise IOError('file does not exist: ' + nxdl_file)
        validate_NXDL(nxdl_file)
        
        self.title = None
        self.category = None
        self.ns = None

        self.ignoreExtraGroups = False
        self.ignoreExtraFields = False
        self.ignoreExtraAttributes = False
        
        # - - - - - - - - - - - - - - - - - - - -
        def get_boolean(attribute, default):
            t = root.get('ignoreExtraGroups', 'false')
            return t.lower() in ('true', '1', True)
        # - - - - - - - - - - - - - - - - - - - -
        
        # parse the XML content now

        tree = lxml.etree.parse(self.nxdl_file_name)
        root = tree.getroot()

        NXDL_mixin.__init__(self, root)
        self.title = root.get('name')
        
        self.ignoreExtraGroups = get_boolean('ignoreExtraGroups', False)
        self.ignoreExtraFields = get_boolean('ignoreExtraFields', False)
        self.ignoreExtraAttributes = get_boolean('ignoreExtraAttributes', False)
        
        # self.category = {
        #              'base': 'base class',
        #              'application': 'application definition',
        #              'contributed': 'contributed definition',
        #              }[root.attrib["category"]]
        self.category = root.attrib["category"]

        self.get_element_data(root, self.category)
    
    def __str__(self):
        return self.title + ' : ' + self.category
    
    def render(self):
        '''
        '''
        indentation = ' '*2
        t = [self.title,]
        for line in self.render_group(self).splitlines():
            t.append(indentation + line)
        return '\n'.join(t)

    def getSubGroup_NX_class_list(self):
        '''
        list the groups used in this NXDL specification by NX_class name
        ''' 
        return sorted({v.NX_class: None for v in self.groups.values()}.keys())


class NX_attribute(NXDL_mixin):
    '''
    NXDL attribute
    '''
    element = 'attribute'

    def __init__(self, node):
        NXDL_mixin.__init__(self, node)
        if self.name in ('restricted deprecated minOccurs'.split()):
            pass
        self.nx_type = self.get_NX_type(node)
        self.units = self.get_NX_units(node)
        self.enum = []
        for n in node.xpath('nx:enumeration/nx:item', namespaces=self.ns):
            self.enum.append( n.attrib['value'] )

    def __str__(self):
        s = '@' + self.name
        s += ' : ' + self.nx_type
        if len(self.enum):
            s += ' = '
            s += ' | '.join(self.enum)
        return s


class NX_field(NXDL_mixin):
    '''
    NXDL field
    '''
    element = 'field'

    def __init__(self, node, category):
        NXDL_mixin.__init__(self, node)
        self.flexible_name = False
        if category in ('base class',):
            self.minOccurs = node.get('minOccurs', 0)
        else:
            self.minOccurs = node.get('minOccurs', 1)
        self.optional = self.minOccurs in ('0', 0)
        
        self.dims = self.field_dimensions(node)

        self.nx_type = self.get_NX_type(node)
        self.units = self.get_NX_units(node)

        self.enum = []
        for n in node.xpath('nx:enumeration/nx:item', namespaces=self.ns):
            self.enum.append( n.attrib['value'] )

        self.attrs = {}
        for subnode in node.xpath('nx:attribute', namespaces=self.ns):
            obj = NX_attribute(subnode)
            self.attrs[obj.name] = obj

    def __str__(self):
        s = self.name
        s += ' : ' + self.nx_type
        if len(self.dims):
            s += '[' + ', '.join(map(str, self.dims)) + ']'
        if len(self.enum):
            s += ' = ' + ' | '.join(self.enum)
        return s
    
    def field_dimensions( self, parent):
        '''
        '''
        node_list = parent.xpath('nx:dimensions', namespaces=self.ns)
        if len(node_list) != 1:
            return []

        dims = {}
        for subnode in node_list[0].xpath('nx:dim', namespaces=self.ns):
            index = int(subnode.get('index'))
            value = subnode.get('value')
            if not value:
                value = 'ref(%s)' % subnode.get('ref')
            if index == 0:
                # index="0": cannot know to which dimension this applies a priori
                value = '*' + str(value) + '*'
            dims[index] = value
        
        if len(dims) == 0:
            # TODO: read the "rank" attribute and create array of that length with "0" values
            rank = int(node_list[0].get('rank'))
            return ['*' for _ in range(rank)]
        if len(dims) == 1 and min(dims.keys()) == 0:
            '''
                The rank of the ``data`` must satisfy
                ``1 <= dataRank <= NX_MAXRANK=32``.  
                At least one ``dim`` must have length ``n``.
            '''
            pass    # index="0": cannot know to which dimension this applies a priori
        elif min(dims.keys()) != 1 or max(dims.keys()) != len(dims):
            msg = 'dimensions not specified properly: ' + str(dims)
            raise KeyError(msg)
        return [dims[k] for k in sorted(map(int, dims.keys()))]


class NX_group(NXDL_mixin):
    '''
    NXDL group
    '''
    element = 'group'

    def __init__(self, node, category):
        NXDL_mixin.__init__(self, node)
        self.NX_class = node.get('type', None)
        if self.NX_class is None:
            msg = 'group has no type, this is an error, name = ' + self.name
            raise ValueError(msg)

        if category in ('base class', ):
            self.flexible_name = True
        else:
            self.flexible_name = False

        try:
            len(self.name)
        except AttributeError:
            self.flexible_name = True
            self.name = '{' + self.NX_class[2:] + '}'

        if category in ('base class',):
            minOccurs = node.get('minOccurs', 0)
        else:
            minOccurs = node.get('minOccurs', 1)
        self.optional = minOccurs in ('0', 0)
        
        self.get_element_data(node, category)

    def __str__(self):
        s = self.name
        s += ' : ' + self.NX_class
        return s


class NX_link(NXDL_mixin):
    '''
    NXDL link
    '''
    element = 'link'

    def __init__(self, node, category):
        NXDL_mixin.__init__(self, node)
        self.target = node.get('target')
    
    def __str__(self):
        return self.name + ' --> ' + self.target


class NX_symbols(NXDL_mixin):
    '''
    NXDL symbols table
    '''
    element = 'symbols'


def get_NXDL_specifications():
    '''
    return a dictionary of NXDL structures, keyed by NX_class name
    '''
    qset = cache.qsettings()
    pfile = cache.get_pickle_file_name(qset.cache_dir())
    if os.path.exists(pfile):
        sha = qset.getKey('git_sha')
        try:
            nxdl_dict = cache.read_pickle_file(pfile, sha)
        except (AttributeError, ImportError):
            # could not read the pickle file, suggest user update the cache
            nxdl_dict = None
        if nxdl_dict is not None:
            return  nxdl_dict

    basedir = os.path.abspath(os.path.join(os.path.dirname(pfile), __init__.NXDL_CACHE_SUBDIR))

    path_list = [
        os.path.join(basedir, 'base_classes'),
        os.path.join(basedir, 'applications'),
        os.path.join(basedir, 'contributed_definitions'),
    ]
    nxdl_file_list = []
    for path in path_list:
        if not os.path.exists(path):
            raise IOError('no definition available, cannot find ' + path)
        for fname in sorted(os.listdir(path)):
            if fname.endswith('.nxdl.xml'):
                nxdl_file_list.append(os.path.join(path, fname))
    
    nxdl_dict = collections.OrderedDict()
    for nxdl_file_name in nxdl_file_list:
        # k = os.path.basename(nxdl_file_name)
        obj = NXDL_definition(nxdl_file_name)
        nxdl_dict[obj.title] = obj

    return nxdl_dict


def parse_command_line_arguments():
    import argparse
    
    doc = __doc__.strip().splitlines()[0]
    doc += '\n  URL: ' + __url__
    doc += '\n  v' + __init__.__version__
    parser = argparse.ArgumentParser(prog=PROGRAM_NAME, description=doc)

    parser.add_argument('infile', 
                    action='store', 
                    help="NXDL file name")

    parser.add_argument('-v', 
                        '--version', 
                        action='version', 
                        version=__init__.__version__)

    return parser.parse_args()


if __name__ == '__main__':
    print "Start this module using:  python main.py structure ..."
    exit(0)



# -*- coding: iso-8859-1 -*-

'''Python Utilities for NeXus HDF5 files'''

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

__author__    = 'Pete R. Jemian'
__email__     = 'prjemian@gmail.com'
__copyright__ = '2016, Pete R. Jemian'

__package_name__ = 'punx'
__license_url__  = 'http://creativecommons.org/licenses/by/4.0/deed.en_US'
__license__      = 'Creative Commons Attribution 4.0 International Public License (see LICENSE file)'
__description__  = 'Python Utilities for NeXus'
__author_name__  = __author__
__author_email__ = __email__
__url__          = u'http://punx.readthedocs.io'
__download_url__ = u'https://github.com/prjemian/punx/archive/master.zip'
__keywords__     = ['NeXus', 'HDF5']

# used by QSettings to store configuration and user cache
__settings_organization__ = __package_name__
__settings_package__ = __package_name__

__install_requires__ = ['h5py', 'lxml', 'numpy', 
                        'pyRestTable', 
                        # 'PyQt4',  
    # see: http://stackoverflow.com/questions/4628519/is-it-possible-to-require-pyqt-from-setuptools-setup-py
                        ]
__sphinx_mock_list__ = __install_requires__ + ['lxml.etree', ]
__classifiers__ = [
     #'Development Status :: 5 - Production/Stable',
     #'Development Status :: 4 - Beta',
     'Development Status :: 3 - Alpha',
     'Environment :: Console',
     'Intended Audience :: Science/Research',
     'License :: Freely Distributable',
     'License :: Public Domain',
     'Programming Language :: Python',
     'Programming Language :: Python :: 2',
     'Programming Language :: Python :: 2.7',
     'Topic :: Scientific/Engineering',
     'Topic :: Scientific/Engineering :: Astronomy',
     'Topic :: Scientific/Engineering :: Bio-Informatics',
     'Topic :: Scientific/Engineering :: Chemistry',
     'Topic :: Scientific/Engineering :: Information Analysis',
     'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
     'Topic :: Scientific/Engineering :: Mathematics',
     'Topic :: Scientific/Engineering :: Physics',
     'Topic :: Scientific/Engineering :: Visualization',
     'Topic :: Software Development',
     'Topic :: Utilities',
   ]

GITHUB_NXDL_ORGANIZATION        = 'nexusformat'
GITHUB_NXDL_REPOSITORY          = 'definitions'
GITHUB_NXDL_BRANCH              = 'master'
GLOBAL_INI_GROUP                = '___global___'
CACHE_SUBDIR                    = 'cache'
SOURCE_CACHE_SETTINGS_FILENAME  = 'punx.ini'
PICKLE_FILE                     = 'nxdl.p'
NXDL_CACHE_SUBDIR = GITHUB_NXDL_REPOSITORY + '-' + GITHUB_NXDL_BRANCH


import os
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
try:
    from ._version import get_versions
except ValueError:
    from _version import get_versions
__version__ = get_versions()['version']
del get_versions
if on_rtd:
    # special handling for readthedocs.org, remove distracting info
    __version__ = __version__.split('+')[0]
__release__   = __version__




'''
manage the settings file for this application
'''

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------


import datetime
import os

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd:
    from mock_PyQt4 import QtCore
else:
    from PyQt4 import QtCore

import __init__


orgName = __init__.__settings_organization__
appName = __init__.__settings_package__


class QSettingsMixin(object):
    '''
    utility methods
    '''

    def init_global_keys(self):
        self.updateGroupKeys({'file': str(self.fileName())})
        self.updateGroupKeys({'version': '1.0'})
    
    def updateGroupKeys(self, group_dict={}, group=__init__.GLOBAL_INI_GROUP):
        for k, v in sorted(group_dict.items()):
            if self.getKey(group + '/' + k) in ('', None):
                self.setKey(group + '/' + k, v)

    def _keySplit_(self, full_key):
        '''
        split full_key into (group, key) tuple
        
        :param str full_key: either `key` or `group/key`, 
            default group (unspecified) is ``__init__.GLOBAL_INI_GROUP``
        '''
        if len(full_key) == 0:
            raise KeyError, 'must supply a key'
        parts = full_key.split('/')
        if len(parts) > 2:
            raise KeyError, 'too many "/" separators: ' + full_key
        if len(parts) == 1:
            group, key = __init__.GLOBAL_INI_GROUP, str(parts[0])
        elif len(parts) == 2:
            group, key = map(str, parts)
        return group, key
    
    def keyExists(self, key):
        '''does the named key exist?'''
        return key in self.allKeys()

    def getKey(self, key):
        '''
        return the Python value (not a QVariant) of key or None if not found
        
        :raises TypeError: if key is None
        '''
        group, k = self._keySplit_(key)
        if group is None:
            group = __init__.GLOBAL_INI_GROUP
        key = group + '/' + k
        v = self.value(key)
        if isinstance(v, QtCore.QVariant):
            v = v.toPyObject()
        return v
    
    def setKey(self, key, value):
        '''
        set the value of a configuration key, creates the key if it does not exist
        
        :param str key: either `key` or `group/key`
        
        Complement:  self.value(key)  returns value of key
        '''
        group, k = self._keySplit_(key)
        if group is None:
            group = __init__.GLOBAL_INI_GROUP
        self.remove(key)
        self.beginGroup(group)
        self.setValue(k, value)
        self.endGroup()
        if key not in ('written', 'written_gmt'):
            self.updateTimeStamp()
 
    def resetDefaults(self):
        '''
        Reset all application settings to default values.
        '''
        for key in self.allKeys():
            self.remove(key)
        self.init_global_keys()
    
    def updateTimeStamp(self):
        '''date/time file was written'''
        key = 'written'
        # current ISO8601 time in local time
        self.setKey(key, timestamp())
        # current ISO8601 time in GMT, matches format from GitHub
        self.setKey(key+'_gmt', gmt_github_style())
    
    def cache_dir(self):
        '''return the absolute path of the cache directory'''
        return os.path.dirname(str(self.fileName()))


def gmt_github_style():
    '''current ISO8601 time in GMT, matches format from GitHub'''
    ts = str(datetime.datetime.utcnow())
    ts = 'T'.join(ts.split()).split('.')[0] + 'Z'
    return ts


def timestamp():
    '''current ISO8601 time in GMT, matches format from GitHub'''
    ts = str(datetime.datetime.now())
    return ts



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2016, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

'''
Python Utilities for NeXus HDF5 files

main user interface file

.. rubric:: Usage

::

    usage: punx [-h] [-v] {hierarchy,show,structure,update,validate} ...
    
    Python Utilities for NeXus HDF5 files URL: http://punx.readthedocs.io
    v0.0.1+4.gff00892.dirty
    
    optional arguments:
      -h, --help            show this help message and exit
      -v, --version         show program's version number and exit
    
    subcommands:
      valid subcommands
    
      {demo,structure,update,validate}
        demo                validate NeXus  demo file: writer_1_3.hdf5
        hierarchy           TBA: show NeXus base class hierarchy
        show                TBA: show program information (about the cache)
        structure           show structure of HDF5 file
        update              update the local cache of NeXus definitions
        validate            validate a NeXus file

'''

import os
import argparse

import __init__


# :see: https://docs.python.org/2/library/argparse.html#sub-commands
# obvious 1st implementations are h5structure and update


def func_demo(args):
    '''
    show what **punx** can do
    
    .. index:: demo
    '''
#     print 'punx update '
#     args.force = False
#     func_update(args)

    path = os.path.dirname(__file__)
    args.infile = os.path.abspath(os.path.join(path, 'data', 'writer_1_3.hdf5'))

    print 'console> punx validate ' + args.infile
    func_validate(args)

    print 'console> punx structure ' + args.infile
    import h5structure
    mc = h5structure.h5structure(args.infile)
    #    :param bool show_attributes: display attributes in output
    show_attributes=True
    mc.array_items_shown = 5
    print '\n'.join(mc.report(show_attributes))


def func_hierarchy(args):
    url = 'http://punx.readthedocs.io/en/latest/analyze.html'
    print 'A chart of the NeXus is in the **punx** documentation.'
    print 'see: ' + url


def func_show(args):
    print 'still in development -- not implemented yet'
    print args


def func_structure(args):
    if args.infile.endswith('.nxdl.xml'):
        import nxdlstructure
        nxdl = nxdlstructure.NXDL_definition(args.infile)
        print nxdl.render()
    else:
        import h5structure
        
        #    :param int limit: maximum number of array items to be shown (default = 5)
        limit=5
        #    :param bool show_attributes: display attributes in output
        show_attributes=True
        
        mc = h5structure.h5structure(os.path.abspath(args.infile))
        mc.array_items_shown = limit
        print '\n'.join(mc.report(show_attributes) or '')


def func_update(args):
    import cache
    cache.update_NXDL_Cache(force_update=args.force)


def func_validate(args):
    import validate
    if args.infile.endswith('.nxdl.xml'):
        result = validate.validate_xml(args.infile)
        if result is None:
            print args.infile, ' validates'
    else:
        import finding
        try:
            validator = validate.Data_File_Validator(args.infile)
        except IOError, _exc:
            print 'File not found: ' + args.infile
            exit(1)
        validator.validate()

        # report the findings from the validation
        #  finding.SHOW_ALL        finding.SHOW_NOT_OK        finding.SHOW_ERRORS
        show_these = finding.SHOW_ALL
        print 'Validation findings'
        print ':file: ' + os.path.basename(validator.fname)
        print ':validation results shown: ', ', '.join(sorted(map(str, show_these)))
        print validator.report_findings(show_these)
        
        print 'summary statistics'
        print validator.report_findings_summary()



def parse_command_line_arguments():
    ''' '''
    doc = __doc__.strip().splitlines()[0]
    doc += '\n  URL: ' + __init__.__url__
    doc += '\n  v' + __init__.__version__
    parser = argparse.ArgumentParser(prog=__init__.__package_name__, 
                                     description=doc,
                                     epilog=__init__.__url__)

    parser.add_argument('-v', 
                        '--version', 
                        action='version', 
                        version=__init__.__version__)

    # TODO: stretch goal: GUI for any of this
    # parser.add_argument('-g', 
    #                     '--gui', 
    #                     help='graphical user interface (TBA)')

    subparsers = parser.add_subparsers(title='subcommands',
                                       description='valid subcommands',)
    
    parser_demo = subparsers.add_parser('demo',
                                        help='demonstrate HDF5 file validation')
    parser_demo.set_defaults(func=func_demo)
    
    # TODO: stretch goal: permit the first two char (or more) of each subcommand to be accepted
    # ?? http://stackoverflow.com/questions/4114996/python-argparse-nargs-or-depending-on-prior-argument?rq=1

    parser_hierarchy = subparsers.add_parser('hierarchy', 
                                             help='show NeXus base class hierarchy')
    parser_hierarchy.set_defaults(func=func_hierarchy)
    #parser_hierarchy.add_argument('something', type=bool, help='something help')
    
#     parser_show = subparsers.add_parser('show', 
#                                         help='show program information (about the cache)')
#     parser_show.set_defaults(func=func_show)
#     parser_show.add_argument('details', type=bool, help='details help')
    
    parser_structure = subparsers.add_parser('structure', 
                                             help='show structure of HDF5 or NXDL file')
    parser_structure.set_defaults(func=func_structure)
    parser_structure.add_argument('infile', help="HDF5 or NXDL file name")
    parser_structure.add_argument('-a', 
                        action='store_false', 
                        default=True,
                        dest='show_attributes',
                        help='Do not print attributes of HDF5 file structure')
    
    parser_update = subparsers.add_parser('update', 
                                          help='update the local cache of NeXus definitions')
    parser_update.set_defaults(func=func_update)
    parser_update.add_argument('-f', '--force', 
                               action='store_true', 
                               default=False, 
                               help='force update (if GitHub available)')
    
    parser_validate = subparsers.add_parser('validate', 
                                            help='validate a NeXus file')
    parser_validate.add_argument('infile', help="HDF5 or NXDL file name")
    parser_validate.set_defaults(func=func_validate)

    return parser.parse_args()


def main():
    ''' '''
    args = parse_command_line_arguments()
    args.func(args)


if __name__ == '__main__':
    main()
